import calendar
import logging
import os
import subprocess
import sys
import time

from elasticsearch import Elasticsearch, NotFoundError
from elastic import get_es_hosts, test_snapshot_repo
import redis
import requests
from utils import execute, get_file_size, exit_if_not_on_server

from squirro.common import config as cfg

try:
    import ujson as json
except ImportError:
    import json

KNOWN_DATABASES = [
    'bing',
    'bulk',
    'configuration',
    'dataloader',
    'diffbot',
    'emailsender',
    'facebook',
    'feed',
    'fileimport',
    'filtering',
    'fingerprint',
    'instagram',
    'machinelearning',
    'plumber',
    'reddit',
    'scheduler',
    'sourcer',
    'squirrel',
    'topic',
    'trends',
    'user',
    'youtube',
]

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("urllib3.util.retry").setLevel(logging.WARNING)

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

def add_cluster_restore_args(subparsers):

    p = subparsers.add_parser('restore', help='Restore cluster from snapshot')
    p.set_defaults(func=restore)
    p.add_argument('--source-path', required=True,
                   help='Target path of the snapshot')
    p.add_argument('--elasticsearch-repo', type=unicode, default='squirro',
                   help='Name of the Elasticsearch repo, defaults to squirro')
    p.add_argument('--keep-closed-indices-days', type=int, default=0,
                   help="Optional. If set, closed indexes older than this number of days " \
                   "will be deleted")

def restore(source_path, elasticsearch_repo, keep_closed_indices_days, **kwargs):

    exit_if_not_on_server()

    #get the last backup label
    es_repo_path = "{0}/elasticsearch".format(source_path)

    status_file_path = "{}/replication_status.json".format(source_path)
    try:
        with open(status_file_path, 'r') as status_file:
            status_data = json.loads(status_file.read())
    except Exception:
        log.error('Failed to load status file %r', status_file_path)
        sys.exit(1)

    if len(status_data) == 0:
        log.error('status file contains no data: %r', status_file_path)

    #get the last label
    status_entry = status_data[0]

    label = status_entry['label']
    created_by = status_entry['created_by']

    log.info('Using snapshot "%s", created by "%s"', label, created_by)

    #get the cluster leader
    cluster_leader = get_cluster_leader_hostname()

    if not cluster_leader:
        cluster_leader_hostname = 'localhost'

    #Elastic
    log.info('Connecting to elasticsearch cluster')
    es = Elasticsearch(hosts=get_es_hosts(), timeout=3600)

    log.info('Verify if Snapshot repository %s is viable', elasticsearch_repo)
    test_snapshot_repo(elasticsearch_repo, es, es_repo_path)

    log.info('Testing if Snapshot %s is available and ready', label)
    snapshot_status = es.snapshot.status(
        repository=elasticsearch_repo,
        snapshot=label)

    state = snapshot_status['snapshots'][0].get('state')

    if state != "SUCCESS":
        log.error("ES Snapshot %s state is %s, cannot proceed", label, state)
        sys.exit(1)
    else:
        log.info("ES Snapshot %s state is %s", label, state)

    log.info('----------------------------------------------')
    log.info('Preflight checks successful, initiate restore.')
    log.info('----------------------------------------------')

    #close all indices
    for index_name in snapshot_status['snapshots'][0]['indices'].iterkeys():
        log.info('Closing ES index %s', index_name)
        try:
            es.indices.close(index=index_name)
        except NotFoundError:
            log.info('Index %s does not exist', index_name)

    output = es.snapshot.restore(
        repository=elasticsearch_repo,
        snapshot=label)

    if not output.get('accepted'):
        log.error('Restore job was rejected by the cluster: %r', output)
        sys.exit(1)
    else:
        log.info('Restore initialized')

    es_health = es.cluster.health()
    es_status = es_health.get('status', 'unknown')
    attempts = 0
    max_attempts = 7200
    attempt_timeout = 5

    while es_status != "green":
        if attempts == max_attempts:
            log.error('Cluster is not in a healthy state, aborting!')
            log.error('Detailed cluster state info: \n %s', json.dumps(es_health, indent=4))
            sys.exit(1)

        log.warning('Cluster state is %s', es_status)
        log.debug('Detailed cluster state info: \n %s', json.dumps(es_health, indent=4))
        time.sleep(attempt_timeout)
        attempts += 1
        es_health = es.cluster.health()
        es_status = es_health.get('status', 'unknown')

    #delete obsolete indices, ones that stay closed after a restore
    if keep_closed_indices_days > 0:
        log.info('Testing if any closed indices are older than %i days...',
                 keep_closed_indices_days)
        es_state_output = es.cluster.state()
        indices = es_state_output['metadata']['indices']

        now = calendar.timegm(time.gmtime()) * 1000
        time_delta = keep_closed_indices_days * 86400 * 1000

        for index, values in indices.iteritems():
            if values['state'] == 'close':
                index_creation_date = int(values['settings']['index']['creation_date'])
                index_age = now - index_creation_date

                if index_age > time_delta:
                    log.info('Deleting outdated, closed index %r', index)
                    es.indices.delete(index=index)

    #restore all databases
    log.info('Restoring all MySQL databases')
    databases = status_entry['databases']

    for db in databases:
        import_mysql(db['database'], cluster_leader_hostname, db['path'])

    log.info('Flush all caches')
    flush_cache()

    log.info('---------------------------------------------------')
    log.info('Snapshot %s successfully restored.', label)
    log.info('---------------------------------------------------')


def import_mysql(db_name, host, input_path):
    """Impots a given path to a given database"""

    #get the db user and password
    config_file = '/etc/squirro/{0}.ini'.format(db_name)

    db_string = ""

    with open(config_file, mode='rb') as f:
        for line in f:
            if line.strip().find('db =') == 0:
                db_string = line.strip()
                break

    if db_string == "":
        log.error('no db config found in %s', config_file)
        return False

    parts_1 = db_string.split('mysql://')[1]
    parts_2 = parts_1.split('@localhost')[0]
    parts_3 = parts_2.split(':')

    #username = parts_3[0]
    password = parts_3[1]

    #create tmp. credentials file
    defaults_file = "/tmp/.my-bcp.cnf"

    with open(defaults_file, mode="wb") as f:
        f.write("[client]\npassword = {0}\n".format(password))

    command_parts = [
        'cat',
        input_path,
        '|',
        'mysql',
        '--defaults-file="{0}"'.format(defaults_file),
        '-u',
        db_name,
        '-h',
        host,
        db_name
    ]

    command = ' '.join(command_parts)
    log.debug('Executing command: %s', command)
    output = execute(command)

    #remove tmp. file
    os.remove(defaults_file)

    if not output and output != '':
        log.error('The MySQL import of database %s has failed, aborting', db_name)
        return False

    log.info('MySQL datbase %s imported from %s (%s)', db_name, input_path,
             get_file_size(input_path))

    return

def get_cluster_leader_hostname():
    """
    Query the cluster service to indentify the leader
    Return hostname on success, otherwise False
    """

    #detect if clusterd is running
    ps_table = subprocess.check_output(['ps', '-A'])

    if 'clusterd' not in ps_table:
        log.error("Cluster Service is not running")
        return False

    #locating the cluster leader
    r = requests.get("http://localhost:81/service/cluster/v0/leader/cluster.json")

    if r.status_code != 200:
        log.error("Cluster Service returned unexpected response code %s, content: %r",
                  r.status_code, r.text)
        return False

    cluster_state = r.json()
    cluster_data = cluster_state.get('data')

    if not cluster_data:
        log.error("Cluster Service returned unexpected data: %r", r.text)
        return False

    leader = cluster_data.get('electionLeader')

    if not leader:
        log.error("No leader elected: %r", r.text)
        return False

    leader_hostname = leader.split(":")[0]
    log.info("Detected Cluster Leader: %r", leader_hostname)

    return leader_hostname

def flush_cache():
    """
    Flushes the redis cache_flush.
    """

    topic_cfg = cfg.get_config('squirro.api.topic')
    redis_password = topic_cfg.get('redis_cache', 'password')
    endpoint_discovery = topic_cfg.get('redis_cache', 'endpoint_discovery')

    if endpoint_discovery:
        redis_host = get_cluster_leader_hostname()
    else:
        redis_host = 'localhost'

    if not redis_host:
        redis_host = 'localhost'

    redis_client = redis.Redis(host=redis_host, port=6380, password=redis_password)

    info = redis_client.info()

    key_count = 0
    for key, value in info.iteritems():
        if 'db' in key:

            if isinstance(value, dict) and 'keys' in value:
                key_count += value['keys']

    log.info(u'Flushing %i keys (%s)', key_count, info['used_memory_human'])
    log.debug('Redis server info: %s', json.dumps(info, indent=2))
    redis_client.flushall()

    log.info('Done')
