import logging
import os
import shutil
import socket
import subprocess
import sys
import time

from elasticsearch import Elasticsearch, NotFoundError
from elastic import get_es_hosts, test_snapshot_repo
import requests
from utils import execute, get_file_size, exit_if_not_on_server

try:
    import ujson as json
except ImportError:
    import json

KNOWN_DATABASES = [
    'bing',
    'bulk',
    'configuration',
    'dataloader',
    'diffbot',
    'emailsender',
    'facebook',
    'feed',
    'fileimport',
    'filtering',
    'fingerprint',
    'instagram',
    'machinelearning',
    'plumber',
    'reddit',
    'scheduler',
    'sourcer',
    'squirrel',
    'topic',
    'trends',
    'user',
    'youtube',
]

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("urllib3.util.retry").setLevel(logging.WARNING)

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

def add_cluster_snapshot_args(subparsers):

    p = subparsers.add_parser('snapshot', help='Creating a cluster snapshot')
    p.set_defaults(func=snapshot)
    p.add_argument('--target-path', required=True,
                   help='Target path of the snapshot')
    p.add_argument('--elasticsearch-repo', type=unicode, default='squirro',
                   help='Name of the Elasticsearch repo, defaults to squirro')
    p.add_argument('--keep-snapshots', type=int, default=10,
                   help='The number of snapshots to keep, default to 10')

def snapshot(target_path, elasticsearch_repo, keep_snapshots, **kwargs):
    """
    Creates a snapshot of Elasticsearch and MySQL/MariaDB
    """

    exit_if_not_on_server()

    #SQL
    cluster_leader = get_cluster_leader_hostname()

    if not cluster_leader:
        cluster_leader = 'localhost'

    #preflight checks
    if not os.path.exists(target_path):
        log.error(u'Please create target folder %s and ensure all nodes with elasticsearch have ' \
                  'access to this location', target_path)

        sys.exit(1)

    es_repo_path = "{0}/elasticsearch".format(target_path)

    if not os.path.exists(es_repo_path):
        log.error(u'Please create elasticsearch repo path %s and ensure all nodes with ' \
                  'elasticsearch have write access to this location.', target_path)
        log.error(u'Add this location to /etc/elasticsearch/elasticsearch.yml as:')
        log.error(u'path.repo: %s', es_repo_path)
        log.error(u'Do this on all elasticsearch nodes and restart elasticsearch')

        sys.exit(1)

    #dump mysql
    timestamp_str = time.strftime("%Y_%m_%d_%H_%M_%S", time.gmtime())
    snapshot_folder = "{0}/{1}/".format(target_path + "/mysql/",
                                        timestamp_str).replace('//', '/')
    os.makedirs(snapshot_folder)
    log.info('Snapshot folder: %s', snapshot_folder)

    database_infos = []

    for database in KNOWN_DATABASES:
        output_path = "{0}/{1}.sql".format(snapshot_folder, database).replace('//', '/')
        backup_path = export_mysql(database, cluster_leader, output_path)

        if not backup_path:
            continue

        database_info = {}
        database_info['database'] = database
        database_info['path'] = backup_path

        database_infos.append(database_info)

    log.debug('Databases: %s', json.dumps(database_infos, indent=2))

    #Elastic
    log.info('Connecting to elasticsearch cluster')
    es = Elasticsearch(hosts=get_es_hosts(), timeout=3600)

    #test and create if needed the es repo
    test_snapshot_repo(elasticsearch_repo, es, es_repo_path)

    log.info(u'Creating new snapshot: %s', timestamp_str)

    snapshot = es.snapshot.create(
        repository=elasticsearch_repo,
        snapshot=timestamp_str)

    if not snapshot.get('accepted'):
        log.error('Cluster rejected snapshot attempt: %r', snapshot)
        sys.exit(1)

    while True:
        snapshot_status = es.snapshot.status(
            repository=elasticsearch_repo,
            snapshot=timestamp_str)

        state = snapshot_status['snapshots'][0].get('state')
        stats = snapshot_status['snapshots'][0].get('stats')

        size_mb = int(stats['processed_size_in_bytes']/ 1024 / 1024)

        if state == "SUCCESS":
            log.info('Snapshot successfully created. Processed %i mb', size_mb)
            break

        if state == "ABORTED" or state == "PARTIAL":
            log.info('Snapshot failed as: Processed %s. %i mb', state, size_mb)
            sys.exit(1)

        log.info('Snapshot state %s. Processed %i mb', state, size_mb)
        time.sleep(3)

    #write status file to nas
    status_file_path = "{}/replication_status.json".format(target_path)
    try:
        with open(status_file_path, 'r') as status_file:
            status_data = json.loads(status_file.read())
    except Exception:
        status_data = []

    to_delete = []

    #ensure we only keep the last n entries
    if len(status_data) > keep_snapshots:
        to_delete = status_data[keep_snapshots:]
        status_data = status_data[0:keep_snapshots]

    #create the new entry
    entry = {}
    entry['label'] = timestamp_str
    entry['created_by'] = socket.gethostname()
    entry['databases'] = database_infos

    status_data.insert(0, entry)

    with open(status_file_path, 'w') as status_file:
        status_file.write(json.dumps(status_data, indent=4))

    #delete the old snapshots

    known_labels = []

    for entry in status_data:
        known_labels.append(entry['label'])

    repo_data = es.snapshot.get(repository=elasticsearch_repo, snapshot="_all")

    #find stragglers
    for snapshot in repo_data['snapshots']:
        entry = {}
        entry['label'] = snapshot['snapshot']

        if not entry['label'] in known_labels:
            to_delete.append(entry)

    deleted_labels = []

    for entry in to_delete:

        if entry['label'] in deleted_labels:
            continue
        else:
            deleted_labels.append(entry['label'])

        snapshot_folder = "{0}/{1}/".format(target_path + "/mysql/",
                                            entry['label']).replace('//', '/')
        log.info('Deleting folder %s', snapshot_folder)

        try:
            shutil.rmtree(snapshot_folder)
        except OSError:
            log.error('Failed to delete folder %s', snapshot_folder)

        log.info('Deleting ES snapshot %s', entry['label'])

        try:
            snapshot_status = es.snapshot.delete(
                repository=elasticsearch_repo,
                snapshot=entry['label'])
        except NotFoundError:
            log.error('Failed to delete ES snapshot %s', entry['label'])

    log.info('Snapshot successfully created.')


def export_mysql(db_name, host, output_path):
    """Exports a given database to a path"""

    #get the db user and password
    config_file = '/etc/squirro/{0}.ini'.format(db_name)

    db_string = ""

    try:
        with open(config_file, mode='rb') as f:
            for line in f:
                if line.strip().find('db =') == 0:
                    db_string = line.strip()
                    break
    except IOError:
        log.warning(u'Skipping database %s as its config file is not present', db_name)
        return None

    if db_string == "":
        log.error('no db config found in %s', config_file)
        return False

    parts_1 = db_string.split('mysql://')[1]
    parts_2 = parts_1.split('@localhost')[0]
    parts_3 = parts_2.split(':')

    #username = parts_3[0]
    password = parts_3[1]

    #create tmp. credentials file

    defaults_file = "/tmp/.my-bcp.cnf"

    with open(defaults_file, mode="wb") as f:
        f.write("[client]\npassword = {0}\n".format(password))

    command_parts = [
        'mysqldump',
        '--defaults-file="{0}"'.format(defaults_file),
        '-u',
        db_name,
        '-h',
        host,
        '--databases',
        db_name,
        '>',
        output_path
    ]

    command = ' '.join(command_parts)
    log.debug('Executing command: %s', command)
    output = execute(command)

    #remove tmp. file
    os.remove(defaults_file)

    if not output and output != '':
        log.error('The MySQL export of database %s has failed, aborting', db_name)
        return False

    log.info('MySQL datbase %s exported to %s (%s)', db_name, output_path,
             get_file_size(output_path))

    return output_path

def get_cluster_leader_hostname():
    """
    Query the cluster service to indentify the leader
    Return hostname on success, otherwise False
    """

    #detect if clusterd is running
    ps_table = subprocess.check_output(['ps', '-A'])

    if 'clusterd' not in ps_table:
        log.error("Cluster Service is not running")
        return False

    #locating the cluster leader
    r = requests.get("http://localhost:81/service/cluster/v0/leader/cluster.json")

    if r.status_code != 200:
        log.error("Cluster Service returned unexpected response code %s, content: %r",
                  r.status_code, r.text)
        return False

    cluster_state = r.json()
    cluster_data = cluster_state.get('data')

    if not cluster_data:
        log.error("Cluster Service returned unexpected data: %r", r.text)
        return False

    leader = cluster_data.get('electionLeader')

    if not leader:
        log.error("No leader elected: %r", r.text)
        return False

    leader_hostname = leader.split(":")[0]
    log.info("Detected Cluster Leader: %r", leader_hostname)

    return leader_hostname
