import logging
import os
import sys
import time

from elasticsearch import Elasticsearch, NotFoundError, TransportError

try:
    import ujson as json
except ImportError:
    import json

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("urllib3.util.retry").setLevel(logging.WARNING)

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

def add_es_args(subparsers):

    p = subparsers.add_parser('elasticsearch', help='Commands related to elasticsearch.')
    es_sub = p.add_subparsers(help='sub-command help')

    replicas = es_sub.add_parser('replicas', help='Set index replica level')
    replicas.set_defaults(func=set_replica_level)

    replicas.add_argument(
        '--name', required=False, type=unicode, help='Name of the elasticsearch index')

    replicas.add_argument(
        '--project-id', required=False, type=unicode,
        help='Squirro project id to locate the elasticsearch index. Is overridden by --name ')

    replicas.add_argument(
        '--project-name', required=False, type=unicode,
        help='Squirro project name to locate the elasticsearch index. Is overidden by --name and ' \
             '--project-id')

    replicas.add_argument(
        '--replica-level', required=True, type=int,
        help='Number of replicas, auto adjusts to prevent numbers bigger than number of nodes -1.' \
        ' If set to -1, number is evaluated by number of nodes - 1')

    create_alias = es_sub.add_parser('create_alias', help='Create an alias')
    create_alias.set_defaults(module='elasticsearch_create_alias')

    create_alias.add_argument(
        '--name', required=True, type=unicode, help='Name of the alias')

    create_alias.add_argument('project_ids', metavar='PROJECT_ID', type=unicode, nargs='+',
                              help='Indexes of these project ids to add to the alias')

    flush = es_sub.add_parser('flush', help='Flush and refresh an index')
    flush.set_defaults(module='elasticsearch_flush')


def set_replica_level(replica_level, index_name, project_id, project_name, **kwargs):
    """
    Set replica level of a project index.
    """

    #get index name and warn about syntax overrrides
    if index_name:
        index_name = index_name
        if project_id:
            log.warn('Ignoring --project-id, --index-name overrides it.')
        if project_name:
            log.warn('Ignoring --project-name, --name overrides it.')
    elif project_id or project_name:
        if project_name and project_id:
            log.warn('Ignoring --project-name, --project-id overrides it.')
    else:
        log.error('Please provide one of --name, --project-id, project-name')

    quit()

    log.info('Connecting to elasticsearch cluster')
    es = Elasticsearch(hosts=['locahost:9200'], timeout=60)
    es_index = 'bla'

    #get number of nodes
    #number_of_nodes = health['number_of_nodes']
    #log.info('Node count: %i', number_of_nodes)

    max_replicas = 10

    if replica_level == -1:
        replica_count = max_replicas
    elif replica_level > max_replicas:
        log.warn('Reduced replica count to, cannot be higher than number of nodes-1')
        replica_count = max_replicas
    else:
        replica_count = replica_level

    log.info('Setting replica count to %i', replica_count)

    body = {
        "index" : {
            "number_of_replicas" : replica_level
        }
    }

    es.indices.get_settings(index=es_index)
    es.indices.put_settings(index=es_index, body=body)


def get_es_health(es_client):
    """
    Helper function to get es cluster health
    """

    res = es_client.cluster.health(level='cluster')
    log.debug('Cluster health: %s', json.dumps(res, indent=2))
    return res


def get_es_hosts():
    """
    Parses nginx config to locate a viable es hostname
    Not using nginx here, because it might timeout on longrunning jobs
    """

    es_hosts = []

    with open('/etc/nginx/conf.d/upstream-elastic.inc', mode='rb') as f:

        for line in f:
            if line.find('server') != -1:
                line = line.replace('server ', '').strip().strip(';')
                parts = line.split(':')
                hostname = parts[0].strip()
                port = parts[1].strip()
                es_hosts.append("http://{0}:{1}".format(hostname, port))

    if len(es_hosts) == 1:
        log.info('Found %i elasticsearch host: %r', len(es_hosts), es_hosts)
    else:
        log.info('Found %i elasticsearch hosts: %r', len(es_hosts), es_hosts)

    return es_hosts


def test_snapshot_repo(elasticsearch_repo, es, es_repo_path):
    es_status = 'unknown'
    attempts = 0
    max_attempts = 60
    attempt_timeout = 15

    while attempts < max_attempts:
        attempts += 1
        es_response = get_es_health(es)
        es_status = es_response.get('status')

        if es_status == 'green':
            break

        log.warn(u'Elasticsearch status is %s, sleeping for %i seconds, attempt %i/%i',
                 es_status, attempt_timeout, attempts, max_attempts)

        time.sleep(attempt_timeout)

    log.info(u'Testing if Snapshot repository "%s" is viable', elasticsearch_repo)

    try:
        verify_data = es.snapshot.verify_repository(repository=elasticsearch_repo)
    except NotFoundError:
        log.warning('Elasticsearch repository "%s" does not exist, creating...',
                    elasticsearch_repo)

        body = {
            "type": "fs",
            "settings": {
                "location": es_repo_path
            }
        }

        log.info('%r', body)

        try:
            es.snapshot.create_repository(repository=elasticsearch_repo, body=body)
        except TransportError as error:
            log.error(u'Failed to create repo, most likely reason is that %s is not writable by ' \
                      'user elasticsearch from all elasticsearch instances in the cluster.',
                      es_repo_path)

            log.error(u'HTTP Status Code: %i: %s:%s', error[0], error[1], json.dumps(error[2],
                                                                                     indent=2))
            sys.exit(1)

        verify_data = es.snapshot.verify_repository(repository=elasticsearch_repo)
    except TransportError as error:
        log.error(u'Failed to create repo, most likely reason is that %s is not writable by ' \
                  'user elasticsearch from all elasticsearch instances in the cluster.',
                  es_repo_path)

        log.error(u'HTTP Status Code: %i: %s:%s', error[0], error[1],
                  json.dumps(error[2], indent=2))
        sys.exit(1)

    #testing if all nodes in good shape
    if es_response['number_of_data_nodes'] != len(verify_data['nodes']):
        log.error('Number of nodes with access to snapshot repo differs from number of nodes ' \
                  ' in cluster')
        log.error('Repo verify data: %s', json.dumps(verify_data, indent=2))
        log.error('ES Cluster state: %s', json.dumps(es_response, indent=2))
        sys.exit(0)

    return
