import datetime
import logging
import os
import sys

from project_export_lib.plugin import ProjectExporter
from utils import exit_if_not_on_server

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3.util.retry").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("elasticsearch").setLevel(logging.WARNING)

def add_project_export_args(subparsers):

    p = subparsers.add_parser('export', help='Commands related to squirro project exports')
    p.set_defaults(func=export)
    p.add_argument('--project-id', required=True, help="Squirro Project ID")
    p.add_argument('--all', dest='all_sources', action='store_true',
                   help='If set all entity types are exported')
    p.add_argument('--sources', action='store_true',
                   help='If set data sources are exported')
    p.add_argument('--facets', action='store_true',
                   help='If set data facets are exported')
    p.add_argument('--enrichments', action='store_true',
                   help='If set enrichments are exported')
    p.add_argument('--smartfilters', action='store_true',
                   help='If set smartfilters are exported')
    p.add_argument('--trends', action='store_true',
                   help='If set trends are exported')
    p.add_argument('--dashboards', action='store_true',
                   help='If set dashboards are exported')
    p.add_argument('--project-configuration', action='store_true',
                   help='If set project configurations are exported')
    p.add_argument('--data', action='store_true',
                   help='If set the squirro items are exported')
    p.add_argument('--archive-path',
                   help='Optional: filename of the archive, should end in .sqproj')
    p.add_argument('--force', action='store_true',
                   help='If set existing export archives are overwritten')

def export(project_id=None, squirro_client=None, all_sources=False, sources=False, facets=False,
           enrichments=False, smartfilters=False, trends=False, dashboards=False,
           project_configuration=False, data=False, archive_path=None, force=False, **kwargs):

    exit_if_not_on_server()

    if not squirro_client:
        log.error('Please set --cluster and --token')
        quit(1)

    project = squirro_client.get_project(project_id)
    log.info('Exporting project %s (id: %s)', project['title'], project['id'])

    options = {}
    if all_sources or sources:
        options['export_sources'] = 'true'

    if all_sources or facets:
        options['export_facets'] = 'true'

    if all_sources or enrichments:
        options['export_enrichments'] = 'true'

    if all_sources or smartfilters:
        options['export_smartfilters'] = 'true'

    if all_sources or trends:
        options['export_trends'] = 'true'

    if all_sources or dashboards:
        options['export_dashboards'] = 'true'

    if all_sources or project_configuration:
        options['export_project_configuration'] = 'true'

    if all_sources or data:
        options['export_data'] = 'true'

    if not len(options):
        log.warn("Nothing to do, please check --help to see the supported entity flags")
        quit(1)

    if archive_path and  archive_path[-1] == "/":
        log.error('Please specify a full --archive-path with a filename that ends in .sqproj')
        quit(1)

    if archive_path:
        if not '.tar.gz' in archive_path and not '.sqproj' in archive_path:
            archive_path = u"{0}.sqproj".format(archive_path)

    if not archive_path:
        #assemble robust archive path
        norm_project_name = project['title'].strip().lower().replace(' ', '_').replace('/', '_').replace('\\', '_').replace('&', '_').replace('__', '_')
        now = datetime.datetime.now()
        date_str = now.strftime("%Y%m%d_%H%M%S")
        archive_path = "/tmp/{0}_{1}.sqproj".format(norm_project_name, date_str)

    if os.path.exists(archive_path) and not force:
        log.warn(u'Path %s exists. Use --force to overwrite it', archive_path)
        quit(1)

    exporter = ProjectExporter(project_id, options, client=squirro_client,
                               archive_path=archive_path)
    archive = exporter.compile_archive()

    log.warn('File exported to %r', archive)
