import base64
import datetime
import json
import logging
import os
import requests
import StringIO
import tarfile
import tempfile
import zipfile

import ujson as json

from flask import make_response, request

from elasticsearch import Elasticsearch
from elasticsearch.helpers import scan
from squirro.common.dependency import get_injected
from squirro.common.config import get_config
from squirro.lib.configuration.project import ProjectConfiguration
from squirro.sdk.studio import StudioPlugin
from squirro_client import NotFoundError

plugin = StudioPlugin(__name__)
log = logging.getLogger(__name__)


ARCHVIVE_NAME = 'archive.tar.gz'
PROJECT_PLACEHOLDER = '$$$SQUIRRO_PROJECT_ID$$$'


@plugin.route('/export', methods=['POST'])
def export():
    project_id = request.form.get('project_id')
    exporter = ProjectExporter(project_id, request.form)
    archive = exporter.compile_archive()

    with open(archive, 'rb') as f:
        res = make_response(f.read())
    res.headers['Content-Disposition'] = 'attachment; filename="' + exporter.filename + '"'
    return res


class ProjectExporter(object):
    def __init__(self, project_id, options, client=None, archive_path=None):

        if client:
            self.client = client
        else:
            self.client = get_injected('squirro_client')
        self.tmpfolder = tempfile.mkdtemp()
        self.project_id = project_id

        if archive_path:
            self.archive_path = archive_path

        bool_options = {}
        for key, value in options.iteritems():
            if value == 'true':
                bool_options[key] = True
            elif value == 'false':
                bool_options[key] = False
            else:
                pass  # Ignore all other values
        self.options = bool_options
        log.info('Exporting project with options: %r', self.options)

    def compile_archive(self):
        self.export_all()
        return self._tar_folder()

    def export_all(self):
        self.export_sources()
        self.export_facets()
        self.export_enrichments()
        self.export_smartfilters()
        self.export_trends()
        self.export_dashboards()
        self.export_project_configuration()
        self.export_data()
        self.export_project()
        self.export_manifest()

    def export_manifest(self):

        #get squirro version info
        topic_cfg = get_config('topic')

        topic_api = topic_cfg.get('services', 'topic')
        version_url = u'{0}/v0/version.json'.format(topic_api)
        log.info('Getting Squirro Version from %r', version_url)
        ret = requests.get(version_url)
        log.info(ret.text)
        version = ret.json()

        data = {
            "version": version["version"],
            "build": version["build"],
            "components": version["components"],

        }

        with self._open('manifest.json', 'wb') as f:
            json.dump(data, f, indent=4)

    def export_project(self):
        data = self.client.get_project(self.project_id)
        with self._open('project.json', 'wb') as f:
            json.dump(data, f, indent=4)

    def export_sources(self):
        data = self.client.get_user_objects(
            self.project_id, full=True, api_version='v1')
        data = data['objects']
        if self.options.get('export_sources'):
            self._dump('sources', data)

        # Find all the used data loader plugins
        if self.options.get('export_dataloader_plugins'):
            plugins = set()
            for obj in data:
                for subscription in obj.get('subscriptions', []):
                    if subscription['provider'] == 'dataloader':
                        plugin = subscription['config']['dataloader_options']['plugin_name']
                        plugins.add(plugin)

            self._export_assets('dataloader_plugin', 'data', plugins)

    def export_data(self):
        #data = self.client.get_dashboards(self.project_id)
        if self.options.get('export_data'):

            #connecting to elasticsearch
            es = Elasticsearch([{'host': 'localhost', 'port': 80, 'url_prefix': '/ext/elastic',
                                 'use_ssl': False}])

            #TODO need to get the reader index from the project locator
            project = self.client.get_project(project_id=self.project_id)
            query={"query": {"match_all": {}}}

            with self._open('data.jsonl', 'wb') as f:
                for index in project['locator']['index_names']:
                    #figure out how many matches there are
                    res = es.count(index=index, doc_type='doc')

                    count = res['count']
                    log.info('Exporting %i docs in index %s', count, index)

                    #estimate the scroll time at 50ms per doc
                    scroll_minutes = count * 50 / 1000 / 60

                    if scroll_minutes < 1:
                        scroll_minutes = 1

                    scan_iterator = scan(es, index=index, doc_type='doc',
                                         scroll='{}m'.format(scroll_minutes), query=query)

                    for res in scan_iterator:
                        f.write(json.dumps(res))
                        f.write('\n')

            log.info('Done!')

    def export_facets(self):
        if self.options.get('export_facets'):
            data = self.client.get_facets(self.project_id)
            self._dump('facets', data)

    def export_enrichments(self):
        data = self.client.get_enrichments(self.project_id)
        if self.options.get('export_enrichments'):
            self._dump('enrichments', data)

        if self.options.get('export_pipelets'):
            #Find all the used pipelets as well
            #TODO need to implement workflow support
            pipelets = set()
            for enrichment in data:
                log.info('%s', json.dumps(enrichment, indent=2))
                if enrichment['type'] == 'pipelet':
                    pipelet = enrichment['config']['pipelet']
                    pipelets.add(pipelet)

            # Export all used pipelets
            pipelets_metadata = []
            for pipelet_id in pipelets:
                pipelet_name = pipelet_id.split('/', 1)[1]
                pipelet_res = self.client.get_pipelet(pipelet_name)
                pipelets_metadata.append(self._export_pipelet(pipelet_res))
            self._dump('enrich/pipelets', pipelets_metadata)

    def export_smartfilters(self):
        if self.options.get('export_smartfilters'):
            smart_filters = self.client.get_project_smartfilters(
                self.project_id, include_config=True)
            ret = []
            for smart_filter in smart_filters:
                if not smart_filter['temporary']:
                    details = self.client.get_smartfilter(
                        'project', self.project_id, smart_filter['name'])
                    ret.append(details)
            self._dump('smartfilters', ret)

    def export_trends(self):
        if self.options.get('export_trends'):
            data = self.client.get_trenddetections(self.project_id)
            data = data['trend_detection_entities']
            self._dump('trends', data)

    def export_dashboards(self):
        data = self.client.get_dashboards(self.project_id)
        if self.options.get('export_dashboards'):
            self._dump('dashboards', data)

        # Find all the used dashboard loaders and widgets
        loaders = set()
        widgets = set()
        for dashboard in data:
            if dashboard.get('loaders'):  # Some legacy dashboards have `False` here
                for loader in dashboard.get('loaders', []):
                    loaders.add(loader)
            for section in dashboard.get('sections', []):
                for widget in section.get('widgets', []):
                    if widget['type'].startswith('custom-'):
                        widget_type = widget['type'].split('-', 1)[1]
                        widgets.add(widget_type)

    def export_project_configuration(self):
        if self.options.get('export_project_configuration'):
            project_config = ProjectConfiguration(self.project_id)
            data = project_config.list()
            if not self.options.get('export_project_configuration_values'):
                for el in data:
                    el['value'] = None
            self._dump('project_configuration', data)

    @property
    def filename(self):
        fname = u'{project}-{date}.sqproj'
        project = self.client.get_project(self.project_id)
        date = datetime.datetime.utcnow().strftime('%Y%m%d')
        return fname.format(project=project['title'], date=date)

    def _dump(self, fname, data):
        if not isinstance(data, list):
            raise Exception(u'Invalid data for {}'.format(fname))
        data = self._clean_project_id(data)
        with self._open(fname + '.json', 'wb') as f:
            json.dump(data, f, indent=4)

        log.info('Exporting %i records to %s', len(data), fname)

    def _clean_project_id(self, data):
        if isinstance(data, basestring) and data == self.project_id:
            return PROJECT_PLACEHOLDER
        elif isinstance(data, list):
            return [self._clean_project_id(el) for el in data]
        elif isinstance(data, dict):
            return {key: self._clean_project_id(value)
                    for key, value in data.iteritems()}
        else:
            return data

    def _get_path(self, fname):
        return os.path.join(self.tmpfolder, fname)

    def _open(self, fname, mode='rb'):
        path = self._get_path(fname)
        folder = os.path.dirname(path)
        if not os.path.exists(folder):
            os.makedirs(folder)
        return open(path, mode)

    def _tar_folder(self):

        if self.archive_path:
            tarfname = self.archive_path
        else:
            tarfname = os.path.join(self.tmpfolder, ARCHVIVE_NAME)

        with tarfile.open(tarfname, 'w:gz') as tar:
            for path in os.listdir(self.tmpfolder):
                if path != ARCHVIVE_NAME:
                    tar.add(os.path.join(self.tmpfolder, path), path)
        return tarfname

    def _export_pipelet(self, pipelet):
        """Write the source and data of the pipelet to disk.

        Return a metadata entry used by the importer.
        """
        pipelet_folder = os.path.join('enrich', pipelet['name'])
        os.makedirs(self._get_path(pipelet_folder))
        source = pipelet.pop('source', None)
        data = pipelet.pop('data', None)
        ret = pipelet

        if source:
            fname = os.path.join(pipelet_folder, pipelet['name'] + '.py')
            with self._open(fname, 'w') as f:
                f.write(source)
            ret['source_file'] = fname

        if data:
            dirname = os.path.join(pipelet_folder, 'data')
            os.mkdir(self._get_path(dirname))
            data = base64.b64decode(data)
            fileobj = StringIO.StringIO(data)
            with zipfile.ZipFile(fileobj) as z:
                z.extractall(self._get_path(dirname))
            ret['data_folder'] = dirname

        return ret

    def _export_assets(self, asset_type, output_path, asset_list):
        for asset in asset_list:
            log.info('asset: %r', asset)
            try:
                asset_obj = self.client.get_asset(asset_type, asset)
            except NotFoundError:
                log.error(u'Failed to export %s asset %r', asset_type, asset)
                continue

            log.debug('Exporting asset: %r', asset_obj)
            asset_path = os.path.join(asset_obj['directory'], asset)
            out_path = os.path.join(self._get_path(output_path), asset)
            for root, dirs, files in os.walk(asset_path):
                relpath = os.path.relpath(root, asset_path)
                out_path_current = os.path.join(out_path, relpath)
                if not os.path.exists(out_path_current):
                    os.makedirs(out_path_current)
                for fname in files:
                    in_fname = os.path.join(root, fname)
                    out_fname = os.path.join(out_path_current, fname)
                    with open(in_fname, 'rb') as inf:
                        with open(out_fname, 'wb') as outf:
                            outf.write(inf.read())
