import logging
import os
import sys
import time
import tarfile

try:
    import ujson as json
except ImportError:
    import json

from project_import_lib.plugin import ProjectImporter
from squirro_client import NotFoundError, UnknownError
from utils import pretty_json

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3.util.retry").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("elasticsearch").setLevel(logging.WARNING)

def add_project_import_args(subparsers):

    p = subparsers.add_parser('import', help='Commands related to squirro project imports')
    p.set_defaults(func=import_project)
    p.add_argument('--project-id', help='Squirro Project ID, if set all data is imported' \
                   'into this project, if not set, then a new project will be created')
    p.add_argument('--restore', action='store_true', help="If set a new project is created for "\
                   "you and all data is " \
                   "imported into it")
    p.add_argument('--force', action='store_true', help="The export command will not overwrite "\
                   " existing project without this flag being set" \
                   "imported into it")
    p.add_argument('--recreate', action='store_true', help="If set the target project will be" \
                   "deleted and recreated prior to the import" \
                   "imported into it")
    p.add_argument('--archive-path', required=True, help="Path to squirro project export archive "\
                   "with suffix .sqproj")

def import_project(archive_path=None, project_id=None, squirro_client=None, restore=False, force=False,
                   recreate=False, **kwargs):

    #preflight checks
    if not project_id and not restore:
        log.error("Set --project-id or --restore")
        quit(1)

    project_exists = False
    target_project = None

    with tarfile.open(archive_path, mode='r:gz') as tar:
        project_file = tar.extractfile('project.json')
        source_project = json.loads(project_file.read())

        if project_id:
            try:
                target_project = squirro_client.get_project(project_id)
                project_exists = True
            except (NotFoundError, UnknownError):
                log.warn('Project ID %s does not exist yet', project_id)
                if not restore:
                    log.info('Create a new project manually or use the --restore flag the creation')
                    quit(1)
        elif restore:
            #test if source project already exists in tenant
            for project in squirro_client.get_projects():
                if project['id'] == source_project['id']:
                    project_exists = True
                    target_project = project
                    break
        else:
            #should not get here
            log.error("Set --project-id or --restore")
            quit(1)

        if project_exists:
            res = squirro_client.query(project_id=project_id, query="*", fields=[], count=1)
            log.debug('project stats: %s', pretty_json(res))
            log.warn('Preflight check:')
            log.warn('%s (id: %s) -> %s (id: %s)', source_project['title'],
                     source_project['id'], target_project['title'], target_project['id'])
            log.warn('The target project contains %i items', res['total'])

            if source_project['title'] != target_project['title']:
                log.warn('NOTE: the project titles are not identical: %s -> %s',
                         source_project['title'], target_project['title'])

            if not force:
                log.error('The project %s (id: %s) already exists on this server. Please use ' \
                          '--force to execute the import.', target_project['title'],
                          target_project['id'])
                log.warn('You can also set --recreate in order to delete and recreate the ' \
                         'project prior to the restore!')
                log.warn('Proceed with care, there is not undo posssible!')
                quit(1)

            log.info(u'Importing project %s into target project "%s"', source_project['title'],
                     target_project['title'])

            log.warn('Press ctrl-c now within 5 seconds to abort...')
            time.sleep(5)

        if recreate:
            if project_exists:
                log.info('Deleting target project %s', target_project['title'])
                squirro_client.delete_project(project_id=target_project['id'])
                time.sleep(5) #for good measure

            try:
                squirro_client.new_project(title=source_project['title'], id=source_project['id'])
            except TypeError:
                log.warn('Recreating project with the same project id is not yet possible in this release')
                target_project = squirro_client.new_project(title=source_project['title'])
                log.warn('Importing into new project with id %s', target_project['id'])

            target_project = squirro_client.get_project(res['id'])

        importer = ProjectImporter(client=squirro_client, project_id=target_project['id'],
                                   tar=tar, callback=log_progress)
        importer.import_all()

def log_progress(msg):
    log.info('%s', msg)
