# encoding: utf8
import inspect
import ujson as json
import logging
import os.path
import shutil
import tempfile

from squirro.lib.configuration.project import ProjectConfiguration
from squirro.sdk.pipelet import upload_pipelet_in_memory
from squirro.sdk.studio import StudioPlugin
from squirro_client import NotFoundError

from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

plugin = StudioPlugin(__name__)
log = logging.getLogger(__name__)

PROJECT_PLACEHOLDER = '$$$SQUIRRO_PROJECT_ID$$$'


# Workaround for SQ-8239
RESERVED_FACET_NAMES = [
    'author',
    'is',
    'language',
    'provider',
    'smartfilter',
    'sort',
    'source',
    'time_increment',
    'read',
    'starred'
]




class ProjectImporter(object):
    # The order here matters so that ID fixup works correctly.
    #
    # Example: trends are referenced by dashboards. So to map the trend ID
    # correctly from the source system to the target system, trends have to be
    # imported first.
    IMPORT_STEPS = [
        # (step, singular, plural, [data_path])
        # `data_path` is an optional file or folder name where the data is
        # located.
        ('project_configuration', 'configuration key', 'configuration keys'),
        ('smartfilters', 'Smart Filter', 'Smart Filters'),
        ('trends', 'trend', 'trends'),
        ('widgets', 'widget', 'widgets', 'dashboard/widgets'),
        ('loaders', 'loader', 'loaders', 'dashboard/loaders'),
        ('dashboards', 'dashboard', 'dashboards'),
        ('pipelets', 'pipelet', 'pipelets', 'enrich/pipelets.json'),
        ('enrichments', 'enrichment', 'enrichments'),
        ('facets', 'facet', 'facets'),
        ('dataloader_plugins', 'data loader plugin', 'data loader plugins', 'data'),
        ('sources', 'source', 'sources'),
        ('data', 'item', 'items', 'data.jsonl'),

    ]

    def __init__(self, client, project_id, tar, callback):
        self.client = client
        self.project_id = project_id
        self.tar = tar
        self.log_callback = callback
        self.import_steps = list(self.IMPORT_STEPS)
        self._id_mapping = {}
        self._assoc_mapping = {}

    def import_all(self):
        total_steps = len(self.import_steps)

        for idx, args in enumerate(self.import_steps):
            if len(args) == 3:
                step, singular, plural = args
                data_path = step + '.json'
            else:
                step, singular, plural, data_path = args

            execute_step = True
            try:
                data_object = self.tar.getmember(data_path)
            except KeyError:
                execute_step = False
                data_object = None

            if execute_step:
                self.log(u'%d/%d Importing %s', idx + 1, total_steps, plural)
                if data_object.isfile():
                    if step == 'data':
                        data = self._get_content(data_path, file_obj=True)
                    else:
                        data = self._get_content(data_path)
                else:
                    data = data_path
                method = getattr(self, 'import_' + step)
                count = method(data)
                if count is None:
                    count = len(data)
                if count == 0:
                    msg = singular
                else:
                    msg = plural
                self.log('%d/%d Imported %d %s', idx + 1, total_steps, count,
                    msg)
            else:
                self.debug('%d/%d Skipping (no data) %s', idx + 1, total_steps,
                    plural)

    def import_project_configuration(self, data):
        project_config = ProjectConfiguration(self.project_id)
        for cfg in data:
            if cfg.get('value') is not None:
                # Always overwrite when value is included in import file
                project_config.set(cfg['key'], cfg)
            elif project_config.get(cfg['key']) is None:
                # Set config when it doesn't yet exist
                project_config.set(cfg['key'], cfg)

    def import_widgets(self, data):
        self.debug('Importing widgets: %r', data)
        return self._import_assets('widget', data, 'widget.ini')

    def import_loaders(self, data):
        self.debug('Importing loaders: %r', data)
        return self._import_assets('dashboard_loader', data, 'loader.js')

    def import_dashboards(self, data):
        for dashboard in data:
            # Themes no longer supported
            if 'theme_id' in dashboard:
                del dashboard['theme_id']
            self._import_entity('dashboard', 'id', 'title', dashboard)

    def import_trends(self, data):
        for trend in data:
            self._import_entity('trenddetection', 'id', 'name', trend,
                get_result_key='trend_detection_entity',
                list_result_key='trend_detection_entities')

    def import_smartfilters(self, data):
        # Custom create_func and modify_func because the Smart Filter methods
        # have a different method signature than other functions.
        def modify_func(project_id, smartfilter_name, **kwargs):
            tmp_sf = self.client.new_smartfilter('project', type_id=project_id)
            tmp_name = tmp_sf['name']
            kwargs.pop('name', None)
            self.client.update_fingerprint_attributes(
                'project', project_id, tmp_name, kwargs)

            for rectype, records in kwargs.get('records', {}).iteritems():
                recs_by_type = {}
                for record in records:
                    recs_by_type.setdefault(record['source'], []).append(record)

                if recs_by_type.get('content'):
                    self.client.update_smartfilter_from_content(
                        'project', project_id, tmp_name,
                        recs_by_type['content'])

            self.client.move_smartfilter('project', project_id, tmp_name,
                smartfilter_name)

            if kwargs.get('locked'):
                self.client.protect_fingerprint(
                    'project', project_id, smartfilter_name, True)

            return tmp_sf

        def create_func(project_id, **kwargs):
            name = kwargs.pop('name')
            return modify_func(project_id, name, **kwargs)

        for sf in data:
            self._import_entity('smartfilter', 'name', 'title', sf,
                get_func=lambda project_id, name: self.client.get_smartfilter(
                    'project', project_id, name),
                list_func=self.client.get_project_smartfilters,
                create_func=create_func, modify_func=modify_func,
                fixup_post_arguments=False)

    def import_enrichments(self, data):
        # Custom `get_func` due to SQ-8229
        def get_func(project_id, enrichment_id):
            enrichments = self.client.get_enrichments(project_id)
            for enrichment in enrichments:
                if enrichment['id'] == enrichment_id:
                    return enrichment
            raise NotFoundError('Enrichment %s not found'.format(
                enrichment_id))

        for enrichment in data:
            self._import_entity('enrichment', 'id', ('name', 'type'),
                enrichment, get_func=get_func,
                create_func=self.client.create_enrichment,
                modify_func=self.client.update_enrichment)

    def import_facets(self, data):
        count = 0
        for facet in data:
            if facet['name'] not in RESERVED_FACET_NAMES:
                self._import_entity('facet', 'id', 'name', facet)
                count += 1
        return count

    def import_pipelets(self, data):
        self.debug('importing pipelets: %r', data)

        count = 0

        for pipelet in data:
            if not pipelet.get('source_file'):
                self.debug('Skipping pipelet %s due to missing source_file',
                     pipelet['name'])
            else:
                content = self._get_content(pipelet['source_file'], is_json=False)
                data = {}
                if pipelet.get('data_folder'):
                    data_folder = pipelet['data_folder'].rstrip('/') + '/'
                    for fname in self.tar.getnames():
                        if fname.startswith(data_folder):
                            d = self._get_content(fname, is_json=False)
                            if d is not None:
                                data[fname] = d

                ret = upload_pipelet_in_memory(self.client, pipelet['name'],
                    content, data)
                self.debug('pipelet uploaded: %r', ret)
                self._id_mapping[pipelet['id']] = ret
                count += 1

    def import_dataloader_plugins(self, data):
        return self._import_assets('dataloader_plugin', data, 'dataloader_plugin.json')

    def import_sources(self, data):
        count = 0
        for obj in data:
            new_obj = self._import_entity('object', 'id', 'title', obj,
                list_func=self.client.get_user_objects)
            self.debug('object %r imported as %r', obj['title'], new_obj)

            for source in obj['subscriptions']:
                # if source['provider'] == 'bulk':
                #     log.info('Skipping bulk source %s (%r)', source['id'],
                #         source['title'])
                #     continue

                #log.info('source: %s', json.dumps(source, indent=2))
                #if source['provider'] == 'dataloader' and source.get('config') and not source['config'].get('title'):
                    # Fix for inconsistent API - title needs to be provided at
                    # a different place than where it's output by the API.
                    #source['config']['title'] = source['title']

                source['object_id'] = new_obj
                if 'token' in source.get('config', {}).get('dataloader_options', {}):
                    source['config']['dataloader_options']['token'] = \
                        self.client.refresh_token

                new_id = self._import_entity('source', 'id', 'title', source,
                    list_func=lambda project_id:
                        self.client.get_object_subscriptions(project_id, new_obj),
                    get_func=lambda project_id, sub_id:
                        self.client.get_subscription(project_id, new_obj, sub_id),
                    get_same_func=lambda project_id, entity:
                        self._get_existing_source(project_id, new_obj, entity),
                    create_func=self.client.new_subscription,
                    modify_func=lambda project_id, subscription_id, config:
                        self.client.modify_subscription(project_id, new_obj,
                            subscription_id, config))

                self._assoc_mapping[source['id']] = new_id
                log.info('new id: %s', new_id)
                count += 1
        return count

    def import_data(self, data):
        count = 0

        #connecting to elasticsearch
        es = Elasticsearch([{'host': 'localhost', 'port': 80, 'url_prefix': '/ext/elastic',
                             'use_ssl': False}])

        project = self.client.get_project(project_id=self.project_id)
        index = project['locator']['index_names'][0]

        data_size = 0
        data_size_limit = 50 * 1024 * 1024 #keep batches at ~50MB
        actions = []
        errors = 0

        for line in data:
            data_size += len(line)

            item = json.loads(line)

            del item['_score']
            del item['sort']
            item['_index'] = index
            item['_op_type'] = 'index'

            if not '_id' in item:
                item['_id'] = item['_source']['item_id']

            #patch the subscription_id
            if 'assoc:subscriptions' in item['_source']:
                new_assoc_ids = []

                for assoc in item['_source']['assoc:subscriptions']:
                    new_assoc_id = self._assoc_mapping.get(assoc)

                    if new_assoc_id:
                        new_assoc_ids.append(new_assoc_id)

                if new_assoc_ids:
                    item['_source']['assoc:subscriptions'] = new_assoc_ids
                else:
                    log.warn('skipping item %r, no matching source found', item['_id'])
                    continue

            count += 1

            actions.append(item)

            if len(actions) == 1000 or data_size > data_size_limit:
                errors += self._flush_actions(es, actions, count)
                data_size = 0
                actions = []

        if len(actions):
            self._flush_actions(es, actions, count)

        if errors:
            log.error('ERROR: A total of %i indexing errors occured. Check the elasticsearch log!', errors)

        return count

    def _flush_actions(self, es, actions, count):

        oks, errors = bulk(es, actions, stats_only=True)
        self.log('Indexed %i items so far...', count)

        if errors:
            pass
            #self.log('-> %i indexing error', errors)

        return errors


    def _get_existing_source(self, project_id, object_id, entity):
        """Return source which is equivalent to the imported source.

        Can't use the usual `title` matching, because source titles are far
        from unique.
        """
        def get_comparison_config(source):
            config = {}
            for key, value in source.get('config', {}).iteritems():
                if source['provider'] == 'dataloader':
                    # We de-duplicate based on `dataloader_plugin_options` and
                    # the plugin.
                    if key == 'dataloader_options':
                        config['plugin_name'] = value['plugin_name']
                    elif key == 'dataloader_plugin_options':
                        config[key] = value
                else:
                    config[key] = value
            return json.dumps(config, sort_keys=True)

        config = get_comparison_config(entity)
        entity_keys = [entity['title'], entity['provider'], config]
        sources = self.client.get_object_subscriptions(project_id, object_id)
        for source in sources:
            source_config = get_comparison_config(source)
            source_keys = [source['title'], source['provider'], source_config]
            if source_keys == entity_keys:
                # If a dataloader source is completely unmodified, then don't
                # update it. This avoids having the source re-created.
                if source['provider'] == 'dataloader':
                    config_keys = ['dataloader_options',
                                   'dataloader_plugin_options',
                                   'scheduling_options']
                    old_values = [source.get('config', {}).get(key)
                                  for key in config_keys]
                    new_values = [entity.get('config', {}).get(key)
                                  for key in config_keys]
                    if old_values == new_values:
                        source['_skip'] = True

                return source
        return None

    def debug(self, msg, *args):
        log.debug(msg, *args)

    def log(self, msg, *args):
        if self.log_callback:
            self.log_callback(msg % args)

    def _import_entity(self, entity_type, id_key, name_key, entity,
            get_result_key=None, list_result_key=None,
            get_func=None, get_same_func=None, list_func=None,
            create_func=None, modify_func=None, fixup_post_arguments=True):
        """Generic import function for entities.
        """
        if isinstance(name_key, basestring):
            # name_key is used as a list, so that multiple keys can be compared
            name_key = [name_key]

        entity_name = entity[name_key[0]]
        entity_id = entity[id_key]
        self.debug('Importing %s %r (ID: %r) ...', entity_type, entity_name,
            entity_id)

        old_entity = self._get_entity(entity_type, entity_id, entity,
            name_key, get_result_key=get_result_key,
            list_result_key=list_result_key,
            get_func=get_func, get_same_func=get_same_func, list_func=list_func)
        if old_entity and old_entity.get('_skip'):
            self.debug('Skipping import of entity: %s %r (ID: %r)',
                entity_type, entity_name, entity_id)
        if old_entity and id_key not in old_entity:
            self.debug('Got entity without ID for type %s: %r', entity_type,
                old_entity)
            raise Exception('Got %s without ID' % entity_type)
        if not old_entity or old_entity.get('_fixup'):
            entity = self._fix_ids(entity)

        if not create_func:
            create_func = getattr(self.client, 'new_' + entity_type)
        if not modify_func:
            modify_func = getattr(self.client, 'modify_' + entity_type)

        # Remove all arguments that are not in the whitelist
        if old_entity:
            call_func = modify_func
        else:
            call_func = create_func
        if fixup_post_arguments:
            # Read out the arguments that we can actually pass in
            self.debug('Fixing entity: %r', entity)
            args = inspect.getargspec(call_func)

            #log.info('args %s: %r: id_key: %r', str(call_func), args, id_key)

            whitelist = [e for e in list(args[0])
                         if e not in ('self', 'project_id')]

            entity_args = {
                key: value
                for key, value in entity.iteritems()
                if key in whitelist
            }
            self.debug('After fixup through whitelist %r: %r', whitelist, entity_args)
        else:
            entity_args = entity

        if old_entity:
            self.debug('Updating %s %r (ID %s)', entity_type, entity_name,
                old_entity[id_key])

            if entity_type != 'smartfilter' and id_key in entity_args:
                del entity_args[id_key]

            #TODO updating of sources is broken atm, needs fixing
            if entity_type != 'source':
                call_func(self.project_id, old_entity[id_key], **entity_args)
            new_id = old_entity[id_key]
        else:
            self.debug('Creating %s %r', entity_type, entity_name)
            new_entity = call_func(self.project_id, **entity_args)
            if get_result_key:
                new_entity = new_entity[get_result_key]
            self.debug('Output: %r', new_entity)
            new_id = new_entity[id_key]

        if new_id:
            self._id_mapping[entity[id_key]] = new_id
        return new_id

    def _get_entity(self, entity_type, id, old_entity, name_key, get_result_key,
            list_result_key, get_func, get_same_func, list_func):
        if not get_func:
            get_func = getattr(self.client, 'get_' + entity_type)
        if not list_func:
            list_func = getattr(self.client, 'get_' + entity_type + 's')

        try:
            ret = get_func(self.project_id, id)
            return ret[get_result_key] if get_result_key else ret
        except NotFoundError:
            pass  # Use list lookup

        ret = None
        if get_same_func:
            ret = get_same_func(self.project_id, old_entity)
        else:
            # check if entity with same name is there
            entities = list_func(self.project_id)
            if list_result_key:
                entities = entities[list_result_key]
            for entity in entities:
                if all(entity[k] == old_entity[k] for k in name_key):
                    ret = entity
                    break

        if ret:
            ret['_fixup'] = True
        return ret

    def _fix_ids(self, data):
        """Replace IDs from the old system with the new ones.
        """
        if isinstance(data, basestring) and data == PROJECT_PLACEHOLDER:
            return self.project_id
        elif isinstance(data, basestring) and data in self._id_mapping:
            return self._id_mapping[data]
        elif isinstance(data, list):
            return [self._fix_ids(el) for el in data]
        elif isinstance(data, dict):
            return {key: self._fix_ids(value)
                    for key, value in data.iteritems()}
        else:
            return data

    def _get_content(self, fname, is_json=True, file_obj=False):
        try:
            fobj = self.tar.extractfile(fname)
        except KeyError:
            fobj = None
        if not fobj:
            return None

        if file_obj:
            return fobj

        if is_json:
            data = json.load(fobj)
        else:
            data = fobj.read()
        return data

    def _import_assets(self, asset_type, folder, file_check=None):
        count = 0

        # Extract asset into a temporary folder
        tmpdir = tempfile.mkdtemp()
        members = [m for m in self.tar.getmembers() if m.name.startswith(folder)]
        self.tar.extractall(tmpdir, members)

        # Upload each asset
        folder_abs = os.path.join(tmpdir, folder)
        for assetdir in os.listdir(folder_abs):
            assetdir_abs = os.path.join(folder_abs, assetdir)
            if not os.path.isdir(assetdir_abs):
                continue
            if file_check and not os.path.isfile(os.path.join(assetdir_abs, file_check)):
                self.debug('Skipping %s %s, because %s does not exist',
                    asset_type, assetdir, file_check)
                continue

            self.debug('Importing %s %s', asset_type, assetdir)
            if asset_type == 'widget':
                config = {
                    'directory': assetdir_abs,
                }
                for line in open(os.path.join(assetdir_abs, 'widget.ini')):
                    if line.startswith('#') or line.startswith('['):
                        continue
                    parts = line.split('=', 1)
                    key = parts[0].strip()
                    if len(parts) == 2 and key != 'hash':
                        config[key] = parts[1].strip()
                log.debug('Got widget config: %r', config)
                self.client.new_dashboard_widget(config=config)
            else:
                self.client.new_asset(asset_type=asset_type, folder=assetdir_abs)

            count += 1

        shutil.rmtree(tmpdir, ignore_errors=True)
        return count
