import argparse
import logging
import os
import subprocess
import sys
import time

try:
    import ujson as json
except ImportError:
    import json

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

VERSION = "2.6.0"

def get_arg_parser():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version=VERSION)
    parser.add_argument('--verbose', '-v', action='count',
                        help='Show additional information.')
    parser.add_argument('--cluster',
                        help='Squirro cluster URL')
    parser.add_argument('--token',
                        help='Squirro user token')
    parser.add_argument('--log-file', dest='log_file',
                        help='Log file on disk.')
    parser.add_argument('--delay', type=int, default=0,
                        help='Delay exit by n seconds, default: 0')

    return parser

def setup_logging(args):
    """Set up logging based on the command line options.
    """
    # Set up logging
    fmt = '%(asctime)s %(name)s %(levelname)-8s %(message)s'
    if args.verbose == 1:
        level = logging.INFO
        logging.getLogger(
            'requests.packages.urllib3.connectionpool').setLevel(logging.WARN)
    elif args.verbose >= 2:
        level = logging.DEBUG
    else:
        # default value
        level = logging.WARN
        logging.getLogger(
            'requests.packages.urllib3.connectionpool').setLevel(logging.WARN)

    # configure the logging system
    if args.log_file:
        out_dir = os.path.dirname(args.log_file)
        if out_dir and not os.path.exists(out_dir):
            os.makedirs(out_dir)
        logging.basicConfig(
            filename=args.log_file, filemode='a', level=level, format=fmt)
    else:
        logging.basicConfig(level=level, format=fmt)

    # Log time in UTC
    logging.Formatter.converter = time.gmtime

def pretty_json(obj):
    return json.dumps(obj, indent=2)

def exit_if_not_on_server():
    if not os.path.exists('/opt/rh/python27/enable'):
        log.error('This module is only supported if invoked on a squirro server')
        log.error('Instructions:')
        log.error('1. ssh into the target server')
        log.error('2. source /opt/rh/python27/enable')
        log.error('3. source /opt/squirro/virtualenv/bin/activate')
        log.error('4. optional if squirro_cli not installed: pip install ' \
                  '/path/to/squirro.cli*.whl)')
        log.error('5. Retry the command')
        quit(1)


def execute(command):
    """Execute a command in a subprocess"""

    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)

    # Poll process for new output until finished
    while True:
        nextline = process.stdout.readline()
        if nextline == '' and process.poll() is not None:
            break
        sys.stdout.write(nextline)
        sys.stdout.flush()

    output = process.communicate()[0]
    exitCode = process.returncode

    if exitCode == 0:
        return output

    log.error('Command %s terminated with code %s, output:%r', command, exitCode, output)
    return False

def get_file_size(file_path):
    """
    Helper to get readable file size
    """

    statinfo = os.stat(file_path)
    size = statinfo.st_size
    if size <= 1024:
        return "%i B" % (size)
    if size <= 1024 * 1024:
        return "%i KB" % (int(size / 1024))
    return "%i MB" % (int(size/1024/1024))
